#include <Arduino.h>
#include <Ps3Controller.h>

#include "main.h"
#include "joyMngt.h"

// #define NOJOY

// ugly method to get rid of callbacks without context
static JoyMngt *currentThis = nullptr;

#define TESTBUTTONPRESSED_DIG(_name_) \
	if (Ps3.event.button_down._name_) \
	{                                 \
		buttonStates._name_ = true;   \
		newEvent = true;              \
	}                                 \
	if (Ps3.event.button_up._name_)   \
	{                                 \
		buttonStates._name_ = false;  \
		newEvent = true;              \
	}

#define TESTBUTTONPRESSED_DIG_TIME(_name_) \
	if (Ps3.event.button_down._name_)      \
	{                                      \
		buttonStates._name_.state = true;  \
		newEvent = true;                   \
	}                                      \
	if (Ps3.event.button_up._name_)        \
	{                                      \
		buttonStates._name_.state = false; \
		newEvent = true;                   \
	}

static void onPS3ConnectStatic(void)
{
#ifndef NOJOY
	currentThis->onConnect();
#endif
}
static void notifyStatic(void)
{
#ifndef NOJOY
	currentThis->notify();
#endif
}
void JoyMngt::onConnect()
{

	Serial.println("Joystick connected...");
	// disconnect wifi to avoid conflicts ?
	// TODO

	countLastEventTimeSeconds = 0;
}

void JoyMngt::notify()
{
#ifndef NOJOY
	bool newEvent = false;
	unsigned long now = millis();
	TESTBUTTONPRESSED_DIG_TIME(cross)
	TESTBUTTONPRESSED_DIG_TIME(square)
	TESTBUTTONPRESSED_DIG_TIME(triangle)
	TESTBUTTONPRESSED_DIG_TIME(circle)

	TESTBUTTONPRESSED_DIG(up)
	TESTBUTTONPRESSED_DIG(down)
	TESTBUTTONPRESSED_DIG(left)
	TESTBUTTONPRESSED_DIG(right)

	TESTBUTTONPRESSED_DIG_TIME(l1)
	TESTBUTTONPRESSED_DIG_TIME(r1)
	TESTBUTTONPRESSED_DIG_TIME(l2)
	TESTBUTTONPRESSED_DIG_TIME(r2)

	TESTBUTTONPRESSED_DIG_TIME(l3)
	TESTBUTTONPRESSED_DIG_TIME(r3)

	TESTBUTTONPRESSED_DIG_TIME(select)
	TESTBUTTONPRESSED_DIG_TIME(start)

	TESTBUTTONPRESSED_DIG(ps)

	if (Ps3.data.analog.stick.lx > ANALOG_JOY_DEAD_REGION)
	{
		if (!buttonStates.anaL_right)
		{
			newEvent = true;
		}
		buttonStates.anaL_left = false;
		buttonStates.anaL_right = true;
	}
	else if (Ps3.data.analog.stick.lx < -ANALOG_JOY_DEAD_REGION)
	{
		if (!buttonStates.anaL_left)
		{
			newEvent = true;
		}
		buttonStates.anaL_right = false;
		buttonStates.anaL_left = true;
	}
	else
	{
		if (buttonStates.anaL_right || buttonStates.anaL_left)
		{
			newEvent = true;
		}
		buttonStates.anaL_right = false;
		buttonStates.anaL_left = false;
	}
	if (Ps3.data.analog.stick.lx > ANALOG_JOY_MOUSE_DEAD_REGION ||
		Ps3.data.analog.stick.lx < -ANALOG_JOY_MOUSE_DEAD_REGION)
	{
		int sign = Ps3.data.analog.stick.lx >= 0 ? 1 : -1;
		buttonStates.anaL_horizValue = sign * Ps3.data.analog.stick.lx * Ps3.data.analog.stick.lx / 256;
	}
	else
	{
		buttonStates.anaL_horizValue = 0;
	}

	if (Ps3.data.analog.stick.ly > ANALOG_JOY_DEAD_REGION)
	{
		if (!buttonStates.anaL_down)
		{
			newEvent = true;
		}
		buttonStates.anaL_up = false;
		buttonStates.anaL_down = true;
	}
	else if (Ps3.data.analog.stick.ly < -ANALOG_JOY_DEAD_REGION)
	{
		if (!buttonStates.anaL_up)
		{
			newEvent = true;
		}
		buttonStates.anaL_down = false;
		buttonStates.anaL_up = true;
	}
	else
	{
		if (buttonStates.anaL_up || buttonStates.anaL_down)
		{
			newEvent = true;
		}
		buttonStates.anaL_up = false;
		buttonStates.anaL_down = false;
	}
	if (Ps3.data.analog.stick.ly > ANALOG_JOY_MOUSE_DEAD_REGION ||
		Ps3.data.analog.stick.ly < -ANALOG_JOY_MOUSE_DEAD_REGION)
	{
		int sign = Ps3.data.analog.stick.ly >= 0 ? 1 : -1;
		buttonStates.anaL_vertValue = sign * Ps3.data.analog.stick.ly * Ps3.data.analog.stick.ly / 256;
	}
	else
	{
		buttonStates.anaL_vertValue = 0;
	}

	if (Ps3.data.analog.stick.rx > ANALOG_JOY_DEAD_REGION)
	{
		if (!buttonStates.anaR_right)
		{
			newEvent = true;
		}
		buttonStates.anaR_left = false;
		buttonStates.anaR_right = true;
	}
	else if (Ps3.data.analog.stick.rx < -ANALOG_JOY_DEAD_REGION)
	{
		if (!buttonStates.anaR_left)
		{
			newEvent = true;
		}
		buttonStates.anaR_right = false;
		buttonStates.anaR_left = true;
	}
	else
	{
		if (buttonStates.anaR_right || buttonStates.anaR_left)
		{
			newEvent = true;
		}
		buttonStates.anaR_right = false;
		buttonStates.anaR_left = false;
	}
	if (Ps3.data.analog.stick.ry > ANALOG_JOY_DEAD_REGION)
	{
		if (!buttonStates.anaR_down)
		{
			newEvent = true;
		}
		buttonStates.anaR_up = false;
		buttonStates.anaR_down = true;
	}
	else if (Ps3.data.analog.stick.ry < -ANALOG_JOY_DEAD_REGION)
	{
		if (!buttonStates.anaR_up)
		{
			newEvent = true;
		}
		buttonStates.anaR_down = false;
		buttonStates.anaR_up = true;
	}
	else
	{
		if (buttonStates.anaR_up || buttonStates.anaR_down)
		{
			newEvent = true;
		}
		buttonStates.anaR_up = false;
		buttonStates.anaR_down = false;
	}

	// battery
	int battery = Ps3.data.status.battery;
	if (battery == ps3_status_battery_low)
	{
		lowBat = true;
	}
	// not enough battery : disconnect
	else if (battery == ps3_status_battery_shutdown || battery == ps3_status_battery_dying)
	{
		doDisconnectJoy = true;
	}
	else
	{
		lowBat = false;
	}

	if (ps3JoyConnected)
	{
		if (buttonStates.ps)
		{
			doDisconnectJoy = true;
		}
		else
		{
			joyOutObj->sendJoyStatesToOut(newEvent, &buttonStates);
		}
		if (newEvent)
		{
			countLastEventTimeSeconds = 0;
		}
	}
#endif
}

void JoyMngt::setConfig(Config *conf)
{
	config = conf;
}

JoyMngt::JoyMngt()
{
	currentThis = this;
}

void JoyMngt::setup()
{
	pinMode(BLUELED, OUTPUT);
	digitalWrite(BLUELED, LOW); // blue led off
}

void JoyMngt::setJoyOut(JoyOut *joyout)
{
	joyOutObj = joyout;
}

void JoyMngt::listen()
{
#ifndef NOJOY
	Ps3.attach(notifyStatic);
	Ps3.attachOnConnect(onPS3ConnectStatic);
	Ps3.begin(config->ps3Mac.c_str());
#endif
	lastMillis = millis();
}

void JoyMngt::loop()
{
	unsigned long nowMillis;
	nowMillis = millis();
	// every second
	if ((int)(nowMillis - lastMillis) >= 1000)
	{
		lastMillis += 1000;
		if (ps3JoyConnected)
		{
			countLastEventTimeSeconds++;
			Ps3.setPlayer(config->numPlayer);
		}
	}
	else if ((int)(nowMillis - lastMillis) < 0) // time rolled ?
	{
		lastMillis = nowMillis;
	}

	if (Ps3.isConnected())
	{
		if (!ps3JoyConnected)
		{
			ps3JoyConnected = true;
			joyOutObj->setJoyConnectedState(true);
		}

		if (countLastEventTimeSeconds > config->inactivityTimeoutS)
		{
			doDisconnectJoy = true;
			Serial.println("Disconnect Joystick because of inactivity...");
		}

		if (doDisconnectJoy)
		{
			// the only method to disconnect BT joystick is to reboot !
			doDisconnectJoy = false;
			Serial.println("PS3 disconnected=>reboot");
			config->setNeedReboot();
		}
	}
	else // not connected
	{
		if (ps3JoyConnected)
		{
			Serial.println("PS3 disconnected...");
			ps3JoyConnected = false;
			joyOutObj->setJoyConnectedState(false);
		}
	}
}
