#include <Arduino.h>

#include "main.h"

#include "joyOut.h"

JoyOut::JoyOut()
{
}

void JoyOut::setup()
{
	pinMode(OUTFIRE1, OUTPUT);
	digitalWrite(OUTFIRE1, LOW);
	pinMode(OUTFIRE2, OUTPUT);
	digitalWrite(OUTFIRE2, LOW);
	pinMode(OUTUP, OUTPUT);
	digitalWrite(OUTUP, LOW);
	pinMode(OUTDOWN, OUTPUT);
	digitalWrite(OUTDOWN, LOW);
	pinMode(OUTLEFT, OUTPUT);
	digitalWrite(OUTLEFT, LOW);
	pinMode(OUTRIGHT, OUTPUT);
	digitalWrite(OUTRIGHT, LOW);
}

void JoyOut::loop()
{
	unsigned long nowMillis = millis();

	if ((nowMillis - lastLedBlinkTimeMillis) > 250) // 4/second
	{
		lastLedBlinkTimeMillis = nowMillis;
		bool newJoyLedState=false;
		if (!joyConnectedState)
		{
			newJoyLedState = false;
		}
		else if (mouseMode)
		{
			newJoyLedState = mouseLedState ? false : true;
		}
		else
		{
			newJoyLedState = true;
		}
		if(mouseLedState!=newJoyLedState)
		{
			mouseLedState=newJoyLedState;
			digitalWrite(BLUELED, mouseLedState ? HIGH : LOW);
		}
	}
	// mouse emulation
	if (mouseMode)
	{
		static const unsigned char mlut[4] = {0x0, 0x01, 0x03, 0x02};
		int i, dir;
		unsigned char v;
		int litlMouseX = mouseX / ANALOG_JOY_MOUSE_DIVISOR;
		int litlMouseY = mouseY / ANALOG_JOY_MOUSE_DIVISOR;
		if (litlMouseX != lastMouseX)
		{
			dir = (litlMouseX > lastMouseX) ? 1 : -1;
			lastMouseX += dir;
			v = mlut[lastMouseX & 0x03];
			digitalWrite(OUTDOWN, v & 0x01 ? HIGH : LOW);
			digitalWrite(OUTRIGHT, v & 0x02 ? HIGH : LOW);
		}
		if (litlMouseY != lastMouseY)
		{
			dir = (litlMouseY > lastMouseY) ? 1 : -1;
			lastMouseY += dir;
			v = mlut[lastMouseY & 0x03];
			digitalWrite(OUTUP, v & 0x01 ? HIGH : LOW);
			digitalWrite(OUTLEFT, v & 0x02 ? HIGH : LOW);
		}
	}
}

void JoyOut::stateMachineButtons(Button_t *button, const char *name, const unsigned long now, const SeqEvent_e seqEvent)
{
	int keyLen = 0;

	if (button->seqIdx == -1)
	{
		return;
	}

	String *curSeqStr;
	if (button->state)
	{
		curSeqStr = &config->sequences_on[seqEvent];
	}
	else
	{
		curSeqStr = &config->sequences_off[seqEvent];
	}

	if (!curSeqStr->length() || button->seqIdx >= curSeqStr->length())
	{
		button->seqIdx = -1;
		return;
	}
	const char *stepInSeqStr = &curSeqStr->c_str()[button->seqIdx];

	if (!*stepInSeqStr) // should never happen
	{
		button->seqIdx = -1;
		return;
	}
#if 0
	if (!button->seqIdx)
	{
		Serial.print("Run sequence for button ");
		Serial.print(name);
		Serial.print(" :");
		Serial.println(*curSeqStr);
	}
#endif
	switch (*stepInSeqStr)
	{
	case 'T': // time delay
	{
		int del = atoi(&stepInSeqStr[1]);
		if ((button->eventTimeMs + del) > now)
		{
			return;
		}
		keyLen = 1;
		while (stepInSeqStr[keyLen] >= '0' && stepInSeqStr[keyLen] <= '9')
			keyLen++;
		//Serial.println("DELAY REACHED");
		break;
	}
	case '+': // loop
	{
		button->seqIdx = 0;
		button->eventTimeMs = now;
		//Serial.println("LOOP");
		return;
	}
	case 'F': // fire on
	{
		if (stepInSeqStr[1] == '1') // F1 : fire 1
		{
			if (!disengaged)
			{
				digitalWrite(OUTFIRE1, HIGH);
			}
			//Serial.println("FIRE1 ON");
		}
		else // F2 : fire 2
		{
			if (!disengaged)
			{
				digitalWrite(OUTFIRE2, HIGH);
			}
			//Serial.println("FIRE2 ON");
		}
		keyLen += 2;
		break;
	}
	case 'f': // fire off
	{
		if (stepInSeqStr[1] == '1') // f1 : fire 1
		{
			if (!disengaged)
			{
				digitalWrite(OUTFIRE1, LOW);
			}
			//Serial.println("FIRE1 OFF");
		}
		else // f2 : fire 2
		{
			if (!disengaged)
			{
				digitalWrite(OUTFIRE2, LOW);
			}
			//Serial.println("FIRE2 OFF");
		}
		keyLen += 2;
		break;
	}
	case 'L': // left on
	{
		if (!disengaged)
		{
			digitalWrite(OUTLEFT, HIGH);
		}
		keyLen++;
		//Serial.println("LEFT ON");
		break;
	}
	case 'l': // left off
	{
		if (!disengaged)
		{
			digitalWrite(OUTLEFT, LOW);
		}
		keyLen++;
		//Serial.println("LEFT OFF");
		break;
	}
	case 'R': // right on
	{
		if (!disengaged)
		{
			digitalWrite(OUTRIGHT, HIGH);
		}
		keyLen++;
		//Serial.println("RIGHT ON");
		break;
	}
	case 'r': // right off
	{
		if (!disengaged)
		{
			digitalWrite(OUTRIGHT, LOW);
		}
		keyLen++;
		//Serial.println("RIGHT OFF");
		break;
	}
	case 'U': // up on
	{
		if (!disengaged)
		{
			digitalWrite(OUTUP, HIGH);
		}
		keyLen++;
		//Serial.println("UP ON");
		break;
	}
	case 'u': // up off
	{
		if (!disengaged)
		{
			digitalWrite(OUTUP, LOW);
		}
		keyLen++;
		//Serial.println("UP OFF");
		break;
	}
	case 'D': // down on
	{
		if (!disengaged)
		{
			digitalWrite(OUTDOWN, HIGH);
		}
		keyLen++;
		//Serial.println("DOWN ON");
		break;
	}
	case 'd': // down off
	{
		if (!disengaged)
		{
			digitalWrite(OUTDOWN, LOW);
		}
		keyLen++;
		//Serial.println("DOWN OFF");
		break;
	}
	default:
		keyLen++; // next
		break;	  // nothing to do ?
	}

	if (!keyLen)
	{
		return;
	}
	int l = strlen(stepInSeqStr);
	if (keyLen > l)
		keyLen = l;
	if (keyLen == l)
	{
		button->seqIdx = -1;
		//Serial.println("END SEQ");
	}
	else
	{
		button->seqIdx += keyLen;
	}
	button->eventTimeMs = now;
}

#define CONNECTBUT(_from_, _to_)                                   \
	if (buttonStates->_from_ != lastButtons._from_ && !disengaged) \
	{                                                              \
		digitalWrite(_to_, buttonStates->_from_ ? HIGH : LOW);     \
	}

#define CONNECTBUTSEQ(_name_)                                               \
	if (newEvent && buttonStates->_name_.state != lastButtons._name_.state) \
	{                                                                       \
		buttonStates->_name_.eventTimeMs = now;                             \
		buttonStates->_name_.seqIdx = 0;                                    \
	}                                                                       \
	stateMachineButtons(&buttonStates->_name_, #_name_, now, SEQ_##_name_);

void JoyOut::sendJoyStatesToOut(const bool newEvent, ButtonStates_t *buttonStates)
{
	static ButtonStates_t lastButtons = {};
	unsigned long now = millis(); // used by macros

	if (!mouseMode)
	{
		CONNECTBUTSEQ(cross);
		CONNECTBUTSEQ(circle);
	}
	else // mouse mode
	{
		// cross is left click
		if (newEvent && buttonStates->cross.state != lastButtons.cross.state)
		{
			buttonStates->cross.eventTimeMs = now;
			buttonStates->cross.seqIdx = 0;
			digitalWrite(OUTFIRE1, buttonStates->cross.state ? HIGH : LOW);
		}
		// circle is right click
		if (newEvent && buttonStates->circle.state != lastButtons.circle.state)
		{
			buttonStates->circle.eventTimeMs = now;
			buttonStates->circle.seqIdx = 0;
			digitalWrite(OUTFIRE2, buttonStates->circle.state ? HIGH : LOW);
		}
	}
	CONNECTBUTSEQ(square);
	CONNECTBUTSEQ(triangle);

	CONNECTBUTSEQ(l1);
	CONNECTBUTSEQ(r1);
	CONNECTBUTSEQ(l2);
	CONNECTBUTSEQ(r2);
	CONNECTBUTSEQ(l3);
	CONNECTBUTSEQ(r3);

	CONNECTBUTSEQ(start);
	if (config->mouseCapable)
	{
		if (newEvent && buttonStates->select.state != lastButtons.select.state)
		{
			buttonStates->select.eventTimeMs = now;
			buttonStates->select.seqIdx = 0;
			if (buttonStates->select.state)
			{
				mouseMode = mouseMode ? false : true;
				mouseX = mouseY = 0;
				lastMouseX = lastMouseY = 0;
				sendAllOff();
				digitalWrite(BLUELED, HIGH);
			}
		}
	}
	else
	{
		CONNECTBUTSEQ(select);
	}

	if (!mouseMode)
	{
		CONNECTBUT(anaL_up, OUTUP);
		CONNECTBUT(anaL_down, OUTDOWN);
		CONNECTBUT(anaL_left, OUTLEFT);
		CONNECTBUT(anaL_right, OUTRIGHT);
	}
	else
	{
		// left analogic is mouse pointer
		mouseX += buttonStates->anaL_horizValue;
		mouseY += buttonStates->anaL_vertValue;
	}

	CONNECTBUT(up, OUTUP);
	CONNECTBUT(down, OUTDOWN);
	CONNECTBUT(left, OUTLEFT);
	CONNECTBUT(right, OUTRIGHT);

	memcpy(&lastButtons, buttonStates, sizeof(ButtonStates_t));
}

void JoyOut::setJoyConnectedState(bool state)
{
	joyConnectedState = state;
	sendAllOff();
	mouseMode = false;
	mouseLedState = state;
	digitalWrite(BLUELED, state ? HIGH : LOW); // blue led on/off
}

void JoyOut::sendAllOff()
{
	digitalWrite(OUTFIRE1, LOW);
	digitalWrite(OUTFIRE2, LOW);
	digitalWrite(OUTUP, LOW);
	digitalWrite(OUTDOWN, LOW);
	digitalWrite(OUTLEFT, LOW);
	digitalWrite(OUTRIGHT, LOW);
}

void JoyOut::setConfig(Config *conf)
{
	config = conf;
}

void JoyOut::setDisengaged(bool state)
{
	disengaged = state;
	if (state)
	{
		sendAllOff();
	}
}
