#include <Arduino.h>
#include <ESPAsyncWebServer.h>
#include <ArduinoJson.h>
#include <LITTLEFS.h>

#include "main.h" // here are default conf
#include "joyMngt.h"
#include "joyOut.h"
#include "config.h"
#include "networking.h"

JoyMngt *joy = nullptr;
JoyOut joyOut;
Config config;
Networking net;

bool blueLedAPFlashState=false;
unsigned long blueLedAPFlashTime=0;

static bool lastAmiga5VState = false;

static bool isAmiga5V()
{
	return digitalRead(AMIGA5VDETECT) ? false : true;
}

void setup()
{
	pinMode(AMIGA5VDETECT, INPUT_PULLUP);
	lastAmiga5VState = isAmiga5V();

	Serial.begin(115200);
	delay(1000);
	Serial.println("STARTING....\n");

	config.setup();
	config.setAmiga5VPresence(lastAmiga5VState);
	if (!config.isAp && !config.tmpAp)
	{
		joy = new JoyMngt();
		joy->setup();
	}
	joyOut.setup();
	joyOut.setConfig(&config);

	net.setConfig(&config);
	net.setDefaultSsid(config.getDefaultSsid());
	net.setup();
	net.setAmiga5VPresence(lastAmiga5VState);

	if (joy)
	{
		joy->setConfig(&config);
		joy->setJoyOut(&joyOut);
		// start listening for ps3 joystick connections
		joy->listen();
	}
	else
	{
		// no joy to init blue led output
		pinMode(BLUELED,OUTPUT);
	}

	Serial.println("READY.");
}

void loop()
{
	if (joy)
	{
		joy->loop();
	}
	else
	{
		unsigned long now = millis();
		// if AP, then flash blue led
		if(now>(blueLedAPFlashTime+500)) // 500ms
		{
			blueLedAPFlashTime=now;
			blueLedAPFlashState=blueLedAPFlashState?false:true;
			digitalWrite(BLUELED,blueLedAPFlashState?LOW:HIGH);
		}
	}
	joyOut.loop();

	if (config.isNeedReboot())
	{
		net.disconnectAll();
		delay(1000);
		ESP.restart();
	}
	bool amiga5VState = isAmiga5V();
	if (amiga5VState != lastAmiga5VState)
	{
		Serial.println("amiga5V state changed");
		lastAmiga5VState = amiga5VState;
		net.setAmiga5VPresence(amiga5VState);
		config.setAmiga5VPresence(amiga5VState);
#if 0
		if (amiga5VState)
		{
			joyOut.setDisengaged(false);
		}
		else
		{
			joyOut.setDisengaged(true);
		}
#endif
	}
}
